#define USEROOT

#include<stdio.h>
#include<stdlib.h>
#include<ctype.h>

#include<qapp.h>
#include<qmainwindow.h>
#include<qmenubar.h>
#include<qpopupmenu.h>
#include<qstatusbar.h>
#include<qmessagebox.h>
#include<qwidget.h>
#include<qbitmap.h>
#include<qgroupbox.h>
#include<qpainter.h>
#include<qpointarray.h>
#include<qpen.h>
#include<qkeycode.h>

#include"tie.h"
#include"pixmaps.inc"

#define true TRUE
#define false FALSE

// enables compilation for qt < 1.4
int setMask(const QRegion &) { return 0; }

#include"guitie.h"
#include"guitie.moc"

static char *modehelp[]= {
	"Free Mode: Drag any piece you like anywhere you like",
	"Play Mode: Drag only the selected Piece and \"submit\". ",
	"Network Mode: not yet implemented."
};

Brett::Brett(GuiSpielfeld*p)
	: QWidget(p,"brett")
{
	_gsf=p; _offs=QPoint(30,30); _shadowPos=-1;
}
void Brett::changeUnitSize(int uw,int uh)
{
	_uw=uw; _uh=uh;  
	QCOORD pts[14]; int i, &w=_uw, &h=_uh;
	for (i=0; i<14; i+=2) {
		pts[i]=hexagon[i]*w; pts[i+1]=hexagon[i+1]*h;
	}
	_qpa.setPoints(7,pts); _qrg=QRegion(_qpa);
	
	_feldRect=QRect(_offs.x(),_offs.y(),16*_uw,10*_uh);
	_depotRect=QRect(_offs.x(),_offs.y(),19*_uw,10*_uh);
	_depotRect.moveBy(_feldRect.right()+1+_offs.x(),0);

	_defSize=QSize(max(_depotRect.right(),_feldRect.right())+_offs.x(),
		max(_depotRect.bottom(),_feldRect.bottom())+_offs.x());

	for(i=0; i<27; i++) setPos(i); 

	QBitmap qbm(4*w,2*h,TRUE); QPainter p; 
	p.begin(&qbm); 
		p.setClipRegion(_qrg); p.fillRect(qbm.rect(),color1); 
	p.end();
	_qpm.resize(4*w,2*h); _qpm.setMask(qbm);
	_qsf.resize(_feldRect.size());
	p.begin(&_qpm); 
		p.fillRect(qbm.rect(),white); p.drawPolygon(_qpa);
	p.end();
	p.begin(&_qsf); QFont f; char ch; p.setFont(f); p.setPen(black); 
		p.fillRect(_qsf.rect(),_gsf->_bgcol); 
		for (i=2; i<23; i++) { ch=i+'A';
			if (!isvalid(i)) continue;
			int x=pos2x(i)*3*_uw, y=pos2y(i)*_uh;
			p.drawPixmap(x,y,_qpm);
			p.drawText(x+2*_uw,y+_uh,0,0,AlignCenter|DontClip,&ch,1);
		}
	p.end();
}

void Brett::resizeEvent(QResizeEvent*) {}
void Brett::paintEvent(QPaintEvent*e) {
	QPainter p(this); 
	if (e->rect() != rect()) { p.setClipRect(e->rect()); }
	paintFeld(p);
	paintShadow(p);
}
void Brett::paintFeld(QPainter &p) {
	p.drawPixmap(_offs,_qsf);
}
void Brett::paintShadow(QPainter &p)
{
	if (!isvalid(_shadowPos)) return;
	p.save();
	int x=3*pos2x(_shadowPos)*_uw, y=pos2y(_shadowPos)*_uh;
	x=x+_feldRect.left(); y=y+_feldRect.top();
	QBrush br(yellow,CrossPattern);
	QRegion rg=_qrg; rg.translate(x,y);
	if (p.hasClipping()) { 
		rg=rg.intersect(p.clipRegion());
	}
	p.setClipRegion(rg); p.setRasterOp(XorROP);
	p.fillRect(x,y,4*_uw,2*_uh,br);
	p.restore();
}
void Brett::moveOver(int pos)
{
	if (pos==_shadowPos) return;
	QPainter p(this); 
	paintShadow(p);
	_shadowPos=pos;
	paintShadow(p);
}
int Brett::getPosForPoint(const QPoint &p) 
{
	// quick-hack
	QPoint q=p+QPoint((3*_uw)/2,_uh);
	q=q-_feldRect.topLeft()+QPoint((3*_uw)*2,_uh*4);
	int x=q.x()/(3*_uw)-2;
	int y=(q.y()/_uh + 2 - x)/2 -2;
	int pos= 5*x + y;
	return (y>=0 && y<5 && isvalid(pos)) ? pos : -1 ;
}
void Brett::setPos(int ind)
{
	int x,y, p=_gsf->coll(ind)->pos(), &w=_uw,&h=_uh;
	if (p!=-1) { // am Spielfeld
		x=pos2x(p)*w*3; y=pos2y(p)*h;
		x=x+_feldRect.left(); y=y+_feldRect.top();
	} else { // daneben
		depot2xy(&x,&y,ind); x*=w*3; y*=h; 
		x=x+_depotRect.left(); y=y+_depotRect.top();
	}
#ifdef USEROOT
	_gsf->guicoll(ind)->recreate(this,0,QPoint(x,y),true);
#endif
	_gsf->guicoll(ind)->setGeometry(x,y,4*w,2*h);
}


GuiSpielfeld::GuiSpielfeld(QWidget*p,const char*n)
	: QMainWindow(p,n)
{
	int i; 
	ind2str(_seldigs,-1); _grabWin=NULL; updateScore();

	_brt=new Brett(this); 
	for (i=0; i<27; i++) { 
		_guicoll[i]=new GuiStein(_coll[i],this,getBrett()); 
	}
	_bgcol=QColor("#000070");

	_brt->changeUnitSize(25,43);

	setCentralWidget(_brt);

	// menu 
	QMenuBar *menu=new QMenuBar(this,"menu");
	  QPopupMenu *game=new QPopupMenu(this);
	    QPopupMenu *mode=new QPopupMenu(this); _modeMenu=mode;
	  QPopupMenu *actn=new QPopupMenu(this);

	menu->insertItem( "&Game" , game);
	  game->insertItem( "&New Game" ,this, SLOT(newGame()) );
	  game->insertSeparator();
	  game->insertItem( "&Game Mode" , mode);
	    _freeModeID=
	    mode->insertItem( "&Free Game" , this, SLOT(setFreeMode()));
	    _playModeID=
	    mode->insertItem( "&Single Game" , this, SLOT(setPlayMode()));
	    _multiModeID=
	    mode->insertItem( "&Network Game" , this, SLOT(setMultiMode()));
	  game->insertSeparator();
	  game->insertItem( "&Quit" ,qApp, SLOT(quit()) );
	menu->insertItem( "&Actions" , actn);
	  actn->insertItem( "&Undo" ,this, SLOT(stornoMove()) );
	  actn->insertItem( "&Re-arrange" ,this, SLOT(resetDepot()) );
	  actn->insertItem( "&Submit" ,this, SLOT(submitStein()) );
	menu->insertSeparator();
	menu->insertItem( "no &Help" );

	statusBar();
	setPlayMode();
}
GuiSpielfeld::~GuiSpielfeld()
{
	int i; for (i=0; i<27; i++) { delete _guicoll[i]; _guicoll[i]=NULL; }
}

GuiStein* GuiSpielfeld::at(int p) 
{
	Stein* s=Spielfeld::at(p);
	return (s) ? _guicoll[s->index()] : (GuiStein*)NULL;
}
bool GuiSpielfeld::place(int p,Stein*s)
	{ bool b=Spielfeld::place(p,s); updateScore(); return b; }
bool GuiSpielfeld::remove(int p)
	{ bool b=Spielfeld::remove(p); updateScore(); return b; }
bool GuiSpielfeld::remove(Stein*s)
	{ bool b=Spielfeld::remove(s); updateScore(); return b; }

bool GuiSpielfeld::placeStein(int pos,Stein*s)
{
	if (place(pos,s)) { setPos(s->index()); return true; }
	return false;
}
bool GuiSpielfeld::removeStein(Stein*s)
{
	if (remove(s->pos())) { setPos(s->index()); return true; }
	return false;
}
void GuiSpielfeld::updateScore()
{
	char str[50]; 
	sprintf(str,"Take it easy - Score: %d",count()); 
	setCaption(str); 
}
void GuiSpielfeld::setSel(int i)
{
	if (i == sel()) return;
	int oldsel=sel(); Spielfeld::setSel(i);
	ind2str(_seldigs,i);
	
	if (oldsel>-1) _guicoll[oldsel]->update();
	if (_sel>-1) _guicoll[_sel]->update();
}
void GuiSpielfeld::setSelChar(char ch)
{
	if (getMode()!=FreeMode) return;
	if (ch<'0' || ch>'9') return;
	if (ch=='0') { 
		ind2str(_seldigs,-1); 
	} else {
		_seldigs[val2dir(ch-'0')]=ch;
	}
	setSel(str2ind(_seldigs));
}
void GuiSpielfeld::placeSelTo(char ch) 
{
	if (sel()==-1) return;
	if (ch==' ') { 
		remove(_coll[_sel]); setPos(_sel); 
	} else {
		int p=char2pos(ch); if (!isvalid(p)) return;
		if (!at(p)) {
			remove(_coll[_sel]);
			placeStein(p,_coll[_sel]); 
		}
	}
}  

void GuiSpielfeld::selNextStein() 
{
	if (sel()>-1 && _coll[sel()]->pos()==-1) return;
	int i;
	do {
		i=rand()%27; 
	} while(_coll[i]->pos() != -1);
	setSel(i);
}

void GuiSpielfeld::keyReleaseEvent(QKeyEvent*e) {
	if (e->key()==Key_Shift) {
		if (_grabWin) { _grabWin->doCursor(0); }
	}
}

void GuiSpielfeld::keyPressEvent(QKeyEvent*e) {
	if (e->key()==Key_Shift) {
		if (_grabWin) { _grabWin->doCursor(ShiftButton); }
	}
	char ch; e->accept(); ch=e->ascii();
	if (isdigit(ch)) { setSelChar(ch); return; } 
	if (isalpha(ch)) { placeSelTo(ch); return; } 
	if (e->state() & ControlButton) {
		switch (e->key()) {
		case Key_I: { int i; char str[20];   // pfusch !
			stornoMove();
			printf("enter Value:"); fflush(stdout);
			fgets(str,sizeof(str),stdin);
			sscanf(str,"%d",&i); 
			srand(i); break;
		}
		case Key_Q: case Key_C: 
			stornoMove();
			qApp->quit(); break;
		case Key_R: { int i; 
			stornoMove();
			for (i=0;i<27;i++) { setPos(i); }
			break;
		}
		case Key_N: { int i;
			stornoMove();
			reset(); updateScore();
			for(i=0; i<27; i++) { setPos(i); }
			break;
		}
		default: e->ignore();
		}
		return;
	} else {
		switch (e->key()) {
		case Key_Space: placeSelTo(' '); break;
		case Key_Return: case Key_Enter: 
			submitStein(); break;
		case Key_Escape: stornoMove(); break;
#if 0
#define Key_Begin 0
		case Key_Insert: setSelChar('0'); break; 
		case Key_End: setSelChar('1'); break; 
		case Key_Down: setSelChar('2'); break; 
		case Key_Next: setSelChar('3'); break; 
		case Key_Left: setSelChar('4'); break; 
		case Key_Begin: setSelChar('5'); break; 
		case Key_Right: setSelChar('6'); break; 
		case Key_Home: setSelChar('7'); break; 
		case Key_Up: setSelChar('8'); break; 
		case Key_Prior: setSelChar('9'); break; 
#endif
		default: e->ignore();
		}
		return;
	}
}

void GuiSpielfeld::newGame() { int i;
	stornoMove(); reset(); updateScore();
	for(i=0; i<27; i++) { setPos(i); }
}
void GuiSpielfeld::stornoMove() { 
	if (_grabWin) { _grabWin->stornoMove(); }
}
void GuiSpielfeld::submitStein() { 
	nextStein();
}
void GuiSpielfeld::nextStein() { 
	selNextStein();
}
void GuiSpielfeld::setFreeMode() { 
	_mode=FreeMode;
	updateModeMenu();
}
void GuiSpielfeld::setPlayMode() { 
	// initialize Random, (do newGame()), 
	_mode=PlayMode;
	updateModeMenu();
}
void GuiSpielfeld::setMultiMode() { 
	// do network-stuff
	_mode=MultiMode;
	updateModeMenu();
}
void GuiSpielfeld::updateModeMenu() {
	_modeMenu->setCheckable( TRUE );
	_modeMenu->setItemChecked(_freeModeID, _mode==FreeMode );
	_modeMenu->setItemChecked(_playModeID, _mode==PlayMode );
	_modeMenu->setItemChecked(_multiModeID, _mode==MultiMode );
	statusBar()->message(
		(_mode==FreeMode) ? modehelp[0] :
		(_mode==PlayMode) ? modehelp[1] :
		(_mode==MultiMode) ? modehelp[2] : "error" );
}
void GuiSpielfeld::resetDepot() { int i; 
	for (i=0;i<27;i++) { setPos(i); }
}

GuiStein::GuiStein(Stein*s,GuiSpielfeld*f,QWidget*p)
	: QWidget(p), _s(s), _feld(f)
{ 
	char str[4]; str[3]='\0';
	str[0]='0'+s->value(0); 
	str[1]='0'+s->value(1); 
	str[2]='0'+s->value(2); 
	setName(str); setBackgroundMode(NoBackground);
	_par=_feld->getBrett(); _grabButton=NoButton;
}
const int GuiStein::_rootFlags=
	WStyle_Customize|WStyle_NoBorder|WStyle_Tool;

// black, white, darkGray, gray, lightGray, red, green, blue, cyan, magenta, 
// yellow, darkRed, darkGreen, darkBlue, darkCyan, darkMagenta, darkYellow
void GuiStein::paintEvent(QPaintEvent*)
{
	QPainter p(this);
	p.setBrush(_feld->_bgcol); p.setPen(NoPen);
	//p.drawPolygon(_feld->_qpa);
	p.drawRect(rect());

	p.drawPixmap(0,0,_feld->_pixcoll[name()[1]-'1']);
	p.drawPixmap(0,0,_feld->_pixcoll[name()[2]-'1']);
	p.drawPixmap(0,0,_feld->_pixcoll[name()[0]-'1']);

	if (_s->index() == _feld->sel()) {
		p.drawPixmap(0,0,_feld->_pixcoll.sel());
	}
}

void GuiStein::resizeEvent(QResizeEvent*)
{
	setMask(_par->_qrg); 
	update();
};
void GuiStein::setPos() { _feld->setPos(_s->index()); }

// values are LeftButton, RightButton, MidButton and NoButton. 
void GuiStein::mousePressEvent(QMouseEvent*e)
{
	if (_feld->getMode()!=FreeMode) {
		if (this != _feld->guicoll(_feld->sel())) return;
	}
	_grabButton=e->button(); 
	if (_grabButton==RightButton) {
		if (!_feld->removeStein(_s)) { setPos(); } 
		if (mouseGrabber()==this) { 
			qApp->restoreOverrideCursor(); 
			releaseMouse(); 
		}
		_par->moveOver(-1); _grabButton=NoButton; 
		return;
	}
	if (mouseGrabber()!=this) {
		_oldPos=_s->pos(); _oldXY=pos(); _oldPar=parentWidget();
		_grabPos=e->pos(); _feld->_grabWin=this;
#ifdef USEROOT
		recreate(NULL,_rootFlags,
			(e->globalPos()-_grabPos),true); 
#endif
		raise(); grabMouse(); qApp->setOverrideCursor(ArrowCursor);
		_dragPos=_s->pos();
		_feld->remove(_s); _par->moveOver(_dragPos);
	}
	doCursor(e->state());
}
void GuiStein::mouseReleaseEvent(QMouseEvent*e)
{
	int btn=e->button();
	if (mouseGrabber() != this) return;
	if (_grabButton != btn) return;

	QPoint p=_par->mapFromGlobal(e->globalPos())-_grabPos;

	int pos=_par->getPosForPoint(p); _par->moveOver(-1);
	if (e->state() & ShiftButton) { btn=MidButton; }
	switch (btn) {
	case LeftButton:
		if (pos!=-1) {
			if (!_feld->placeStein(pos,_s)) { setPos(); }
		} else {
			setPos();
		}
		break;
	case MidButton:
		if (pos!=-1) {
			_feld->placeStein(pos,_s);
		} 
		break;
	case RightButton:
		if (!_feld->removeStein(_s)) { setPos(); } 
		break;
	}
	qApp->restoreOverrideCursor(); _dragPos=-1; releaseMouse();
	_feld->_grabWin=NULL;
}
void GuiStein::mouseMoveEvent(QMouseEvent*e)
{
	if (mouseGrabber() != this) return;
#ifdef USEROOT
	move(e->globalPos()-_grabPos);
#else
	move(_par->mapFromGlobal(e->globalPos())-_grabPos);
#endif
	QPoint p=_par->mapFromGlobal(e->globalPos())-_grabPos;
	_dragPos=_par->getPosForPoint(p);
	_par->moveOver(_dragPos);
	doCursor(e->state());
}

void GuiStein::stornoMove()
{
	if (mouseGrabber() != this) return;
	_par->moveOver(-1); _dragPos=-1; _feld->_grabWin=NULL;
	qApp->restoreOverrideCursor(); releaseMouse();
	if (_oldPos != -1) { 
		_feld->placeStein(_oldPos,_s);
	} else {
		if (_oldPar != parentWidget()) {
			recreate(_oldPar,_oldPar?0:_rootFlags,_oldXY,true);
		} else {
			move(_oldXY);
		}
	}
}

void GuiStein::doCursor(int state)
{
        if (mouseGrabber() != this) return;
	if ((_dragPos==-1 || _feld->at(_dragPos)) 
            && !(state & ShiftButton || _grabButton==MidButton)) { 
		 qApp->setOverrideCursor(arrowCursor,TRUE); 
	} else { qApp->setOverrideCursor(crossCursor,TRUE); }
}

int main(int argc,char *argv[]) 
{
	QApplication q(argc,argv);
	GuiSpielfeld *feld=new GuiSpielfeld;

	q.setMainWidget(feld);
	feld->show();
	q.exec();
	delete feld;
}
