#include<stdio.h>
#include<stdlib.h>
#include<ctype.h>

#ifdef DEFBOOL
enum bool {false=0, true=-1 };
#endif

#include"tie.h"

bool isvalid(int pos) { return (bool)(pos>1 && pos!=5 && pos!=19 && pos<23); }
int hexagon[14]={ 1,0 , 3,0 , 4,1 , 3,2 , 1,2 , 0,1 , 1,0 };

int row2val(int i,int j) {
	static int rv[9]={1,5,9 , 3,4,8 , 2,6,7 };
	return rv[i+3*j];
}
int val2dir(int v) {
	static int vd[9]={0,2,1 , 1,0,2 , 2,1,0 };
	return vd[v-1];
}
void ind2str(char *str,int ind)
{
    if (ind==-1) { str[0]=str[1]=str[2]='0'; }
    else {
        str[0]= row2val(ind/9,0) + '0'; 
        str[1]= row2val((ind/3)%3,1) + '0'; 
        str[2]= row2val(ind%3,2) + '0';
    }
}
int str2ind(const char *str)
{
    int s[3]; int i;
    
    for(i=0;i<3;i++) { s[i]=-1; if (str[i]<'1' || str[i]>'9') return -1; }
    for(i=0;i<3;i++) { s[val2dir(str[i]-'0')]=(str[i]-'1')/3; }
    for(i=0;i<3;i++) { if (s[i]<0) return -1; }
    return s[0]*9 + s[1]*3 + s[2];
}

void pos2xy(int *x,int *y,int p) { *x=pos2x(p); *y=pos2y(p); }
void depot2xy(int *x,int *y,int ind) { 
    int d0=ind/9, d1=(ind/3)%3, d2=ind%3, d3=(2*d1+d2)%3;
    if (d0==2) { *x= d3   ; *y= (2*d1+d3)%6; }  else
    if (d0==1) { *x= d3+3 ; *y= (2*d1+d3+3)%6; } else
    if (d0==0) { *x= d3+3*(((2*d1+d3)%6)/3) ; *y= (2*d1+d3)%3+6; } else 
    { *x=*y=0; }
}

Stein::Stein(int ind) : _ind(ind) 
{
	_d[0]=row2val(ind/9,0); _d[1]=row2val((ind/3)%3,1); 
	_d[2]=row2val(ind%3,2); _pos=-1;
}

Spielfeld::Spielfeld() 
{
	int i;  _sel=-1;
	for (i=0;i<27;i++) { _feld[i]=NULL; _coll[i]=new Stein(i); }
}
Spielfeld::~Spielfeld()
{
	int i;
	for (i=0;i<27;i++) { delete _coll[i]; _coll[i]=NULL; _feld[i]=NULL; }
}

int Spielfeld::count() const { int sum=0;
	sum+=rowsum( 2,1,5);  // CHMRW
	sum+=rowsum(10,0,5);  // KLMNO
	sum+=rowsum( 4,2,5);  // EIMQU
	sum+=rowsum( 6,0,4);  // GHIJ
	sum+=rowsum( 6,1,4);  // GLQV
	sum+=rowsum(15,0,4);  // PQRS
	sum+=rowsum( 3,2,4);  // DHLP
	sum+=rowsum( 9,2,4);  // JNRV
	sum+=rowsum( 3,1,4);  // DINS
	sum+=rowsum( 2,2,3);  // CGK
	sum+=rowsum( 2,0,3);  // CDE
	sum+=rowsum( 4,1,3);  // EJO
	sum+=rowsum(10,1,3);  // KPU
	sum+=rowsum(20,0,3);  // UVW
	sum+=rowsum(14,2,3);  // OSW
	return sum;
}
int Spielfeld::rowsum(int pos,int dir,int len) const
{
	//  offset der pos, um in die index-dir zu gehen.
	static const int delta[3]={1,5,4};
	int i,p,dp=delta[dir],tval; Stein *tmp;

	tmp=at(pos); if (!tmp) return 0;
	tval=tmp->value(dir);
	for (i=1,p=pos+dp ; i<len ; i++,p+=dp) {
		tmp=at(p); if (!tmp) return 0;
		if (tmp->value(dir) != tval) return 0;
	}
	return len*tval;
}

bool Spielfeld::place(int p,Stein* s)
{
	if (!s || !isvalid(p) || _feld[p] || s->pos()!=-1) return false;
	_feld[p]=s; s->place(p); return true;
}

bool Spielfeld::remove(int p)
{
	if (!isvalid(p) || _feld[p]==NULL || _feld[p]->pos()==-1) return false;
	_feld[p]->remove(); _feld[p]=NULL; return true;
}

bool Spielfeld::remove(Stein *s) { return remove(s->pos()); }

void Spielfeld::reset()
{
	int i; for (i=0; i<25; i++) { remove(i); }
}

void Spielfeld::textdraw() const
{
	char screen[9][70]; int x,y,i;
	for (y=0;y<9;y++) { 
		for (x=0;x<68;x++) {
			screen[y][x]=' ';
		} screen[y][x]='\n'; screen[y][x+1]='\0';
	}
	fprintf(stdout,"Current value: %d\n",count());
	for (i=0; i<25; i++) { if (!isvalid(i)) continue;
		Stein* tmp=at(i); x=pos2x(i)*6+4; y=pos2y(i);
		char *line=screen[y];
		if (tmp) {
			if (tmp->index()==sel()) { 
				line[x-1]='<'; line[x+3]='>'; 
			}
			ind2str( line+x , tmp->index() );
		} else {
			line[x+1]=pos2char(i);
		}
	}
	for (y=0;y<9;y++) { 
		fputs(screen[y],stdout);
	}
}

